@@ -1,6 +1,6 @@
 """Shared definitions used by different parts of semantic analysis."""
 
-from abc import abstractmethod, abstractproperty
+from abc import abstractmethod
 
 from typing import Optional, List, Callable
 from typing_extensions import Final
@@ -67,7 +67,8 @@ class SemanticAnalyzerCoreInterface:
         """Is a module or class namespace potentially missing some definitions?"""
         raise NotImplementedError
 
-    @abstractproperty
+    @property
+    @abstractmethod
     def final_iteration(self) -> bool:
         """Is this the final iteration of semantic analysis?"""
         raise NotImplementedError
@@ -156,10 +157,15 @@ class SemanticAnalyzerInterface(SemanticAnalyzerCoreInterface):
     def qualified_name(self, n: str) -> str:
         raise NotImplementedError
 
-    @abstractproperty
+    @property
+    @abstractmethod
     def is_typeshed_stub_file(self) -> bool:
         raise NotImplementedError
 
+    @abstractmethod
+    def is_func_scope(self) -> bool:
+        raise NotImplementedError
+
 
 def create_indirect_imported_name(file_node: MypyFile,
                                   module: str,
@@ -208,7 +214,7 @@ def calculate_tuple_fallback(typ: TupleType) -> None:
     Note that there is an apparent chicken and egg problem with respect
     to verifying type arguments against bounds. Verifying bounds might
     require fallbacks, but we might use the bounds to calculate the
-    fallbacks. In partice this is not a problem, since the worst that
+    fallbacks. In practice this is not a problem, since the worst that
     can happen is that we have invalid type argument values, and these
     can happen in later stages as well (they will generate errors, but
     we don't prevent their existence).
