@@ -50,13 +50,23 @@ static void CPyDebug_Print(const char *msg) {
 // we know that it is there.
 static inline CPyVTableItem *CPy_FindTraitVtable(PyTypeObject *trait, CPyVTableItem *vtable) {
     int i;
-    for (i = -2; ; i -= 2) {
+    for (i = -3; ; i -= 3) {
         if ((PyTypeObject *)vtable[i] == trait) {
             return (CPyVTableItem *)vtable[i + 1];
         }
     }
 }
 
+// Use the same logic for offset table.
+static inline size_t CPy_FindAttrOffset(PyTypeObject *trait, CPyVTableItem *vtable, size_t index) {
+    int i;
+    for (i = -3; ; i -= 3) {
+        if ((PyTypeObject *)vtable[i] == trait) {
+            return ((size_t *)vtable[i + 2])[index];
+        }
+    }
+}
+
 static bool _CPy_IsSafeMetaClass(PyTypeObject *metaclass) {
     // mypyc classes can't work with metaclasses in
     // general. Through some various nasty hacks we *do*
@@ -738,6 +748,10 @@ static bool CPySet_Remove(PyObject *set, PyObject *key) {
     return false;
 }
 
+static PyObject *CPyList_Extend(PyObject *o1, PyObject *o2) {
+    return _PyList_Extend((PyListObject *)o1, o2);
+}
+
 static PyObject *CPySequenceTuple_GetItem(PyObject *tuple, CPyTagged index) {
     if (CPyTagged_CheckShort(index)) {
         Py_ssize_t n = CPyTagged_ShortAsSsize_t(index);
@@ -763,6 +777,18 @@ static PyObject *CPySequenceTuple_GetItem(PyObject *tuple, CPyTagged index) {
     }
 }
 
+static PyObject *CPySequence_Multiply(PyObject *seq, CPyTagged t_size) {
+    Py_ssize_t size = CPyTagged_AsSsize_t(t_size);
+    if (size == -1 && PyErr_Occurred()) {
+        return NULL;
+    }
+    return PySequence_Repeat(seq, size);
+}
+
+static PyObject *CPySequence_RMultiply(CPyTagged t_size, PyObject *seq) {
+    return CPySequence_Multiply(seq, t_size);
+}
+
 static CPyTagged CPyObject_Hash(PyObject *o) {
     Py_hash_t h = PyObject_Hash(o);
     if (h == -1) {
@@ -930,6 +956,60 @@ static PyObject *CPyDict_FromAny(PyObject *obj) {
     }
 }
 
+static PyObject *CPyStr_GetItem(PyObject *str, CPyTagged index) {
+    if (PyUnicode_READY(str) != -1) {
+        if (CPyTagged_CheckShort(index)) {
+            Py_ssize_t n = CPyTagged_ShortAsSsize_t(index);
+            Py_ssize_t size = PyUnicode_GET_LENGTH(str);
+            if ((n >= 0 && n >= size) || (n < 0 && n + size < 0)) {
+                PyErr_SetString(PyExc_IndexError, "string index out of range");
+                return NULL;
+            }
+            if (n < 0)
+                n += size;
+            enum PyUnicode_Kind kind = (enum PyUnicode_Kind)PyUnicode_KIND(str);
+            void *data = PyUnicode_DATA(str);
+            Py_UCS4 ch = PyUnicode_READ(kind, data, n);
+            PyObject *unicode = PyUnicode_New(1, ch);
+            if (unicode == NULL)
+                return NULL;
+
+            if (PyUnicode_KIND(unicode) == PyUnicode_1BYTE_KIND) {
+                PyUnicode_1BYTE_DATA(unicode)[0] = (Py_UCS1)ch;
+            }
+            else if (PyUnicode_KIND(unicode) == PyUnicode_2BYTE_KIND) {
+                PyUnicode_2BYTE_DATA(unicode)[0] = (Py_UCS2)ch;
+            } else {
+                assert(PyUnicode_KIND(unicode) == PyUnicode_4BYTE_KIND);
+                PyUnicode_4BYTE_DATA(unicode)[0] = ch;
+            }
+            return unicode;
+        } else {
+            PyErr_SetString(PyExc_IndexError, "string index out of range");
+            return NULL;
+        }
+    } else {
+        PyObject *index_obj = CPyTagged_AsObject(index);
+        return PyObject_GetItem(str, index_obj);
+    }
+}
+
+static PyObject *CPyStr_Split(PyObject *str, PyObject *sep, CPyTagged max_split)
+{
+    Py_ssize_t temp_max_split = CPyTagged_AsSsize_t(max_split);
+    if (temp_max_split == -1 && PyErr_Occurred()) {
+        PyErr_SetString(PyExc_OverflowError, "Python int too large to convert to C ssize_t");
+            return NULL;
+    }
+    return PyUnicode_Split(str, sep, temp_max_split);
+}
+
+/* This does a dodgy attempt to append in place  */
+static PyObject *CPyStr_Append(PyObject *o1, PyObject *o2) {
+    PyUnicode_Append(&o1, o2);
+    return o1;
+}
+
 static PyObject *CPyIter_Next(PyObject *iter)
 {
     return (*iter->ob_type->tp_iternext)(iter);
@@ -994,6 +1074,16 @@ static PyObject *CPyLong_FromFloat(PyObject *o) {
     }
 }
 
+static PyObject *CPyLong_FromStrWithBase(PyObject *o, CPyTagged base) {
+        Py_ssize_t base_size_t = CPyTagged_AsSsize_t(base);
+        return PyLong_FromUnicodeObject(o, base_size_t);
+}
+
+static PyObject *CPyLong_FromStr(PyObject *o) {
+    CPyTagged base = CPyTagged_FromSsize_t(10);
+    return CPyLong_FromStrWithBase(o, base);
+}
+
 // Construct a nicely formatted type name based on __module__ and __name__.
 static PyObject *CPy_GetTypeName(PyObject *type) {
     PyObject *module = NULL, *name = NULL;
@@ -1062,6 +1152,7 @@ static PyObject *CPy_FormatTypeName(PyObject *value) {
     return output;
 }
 
+CPy_NOINLINE
 static void CPy_TypeError(const char *expected, PyObject *value) {
     PyObject *out = CPy_FormatTypeName(value);
     if (out) {
@@ -1162,16 +1253,28 @@ static inline PyObject *_CPy_FromDummy(PyObject *p) {
     return p;
 }
 
-static void CPy_CatchError(PyObject **p_type, PyObject **p_value, PyObject **p_traceback) {
+#ifndef MYPYC_DECLARED_tuple_T3OOO
+#define MYPYC_DECLARED_tuple_T3OOO
+typedef struct tuple_T3OOO {
+    PyObject *f0;
+    PyObject *f1;
+    PyObject *f2;
+} tuple_T3OOO;
+static tuple_T3OOO tuple_undefined_T3OOO = { NULL, NULL, NULL };
+#endif
+
+
+static tuple_T3OOO CPy_CatchError(void) {
     // We need to return the existing sys.exc_info() information, so
     // that it can be restored when we finish handling the error we
     // are catching now. Grab that triple and convert NULL values to
     // the ExcDummy object in order to simplify refcount handling in
     // generated code.
-    PyErr_GetExcInfo(p_type, p_value, p_traceback);
-    _CPy_ToDummy(p_type);
-    _CPy_ToDummy(p_value);
-    _CPy_ToDummy(p_traceback);
+    tuple_T3OOO ret;
+    PyErr_GetExcInfo(&ret.f0, &ret.f1, &ret.f2);
+    _CPy_ToDummy(&ret.f0);
+    _CPy_ToDummy(&ret.f1);
+    _CPy_ToDummy(&ret.f2);
 
     if (!PyErr_Occurred()) {
         PyErr_SetString(PyExc_RuntimeError, "CPy_CatchError called with no error!");
@@ -1193,11 +1296,12 @@ static void CPy_CatchError(PyObject **p_type, PyObject **p_value, PyObject **p_t
     // Clear the error indicator, since the exception isn't
     // propagating anymore.
     PyErr_Clear();
+
+    return ret;
 }
 
-static void CPy_RestoreExcInfo(PyObject *type, PyObject *value, PyObject *traceback) {
-    // PyErr_SetExcInfo steals the references to the values passed to it.
-    PyErr_SetExcInfo(_CPy_FromDummy(type), _CPy_FromDummy(value), _CPy_FromDummy(traceback));
+static void CPy_RestoreExcInfo(tuple_T3OOO info) {
+    PyErr_SetExcInfo(_CPy_FromDummy(info.f0), _CPy_FromDummy(info.f1), _CPy_FromDummy(info.f2));
 }
 
 static void CPy_Raise(PyObject *exc) {
@@ -1218,6 +1322,10 @@ static void CPy_Reraise(void) {
     PyErr_Restore(p_type, p_value, p_traceback);
 }
 
+static int CPy_NoErrOccured(void) {
+    return PyErr_Occurred() == NULL;
+}
+
 static void CPyErr_SetObjectAndTraceback(PyObject *type, PyObject *value, PyObject *traceback) {
     // Set the value and traceback of an error. Because calling
     // PyErr_Restore takes away a reference to each object passed in
@@ -1256,13 +1364,274 @@ static inline void _CPy_ToNone(PyObject **p) {
     }
 }
 
-static void CPy_GetExcInfo(PyObject **p_type, PyObject **p_value, PyObject **p_traceback) {
+static void _CPy_GetExcInfo(PyObject **p_type, PyObject **p_value, PyObject **p_traceback) {
     PyErr_GetExcInfo(p_type, p_value, p_traceback);
     _CPy_ToNone(p_type);
     _CPy_ToNone(p_value);
     _CPy_ToNone(p_traceback);
 }
 
+static tuple_T3OOO CPy_GetExcInfo(void) {
+    tuple_T3OOO ret;
+    _CPy_GetExcInfo(&ret.f0, &ret.f1, &ret.f2);
+    return ret;
+}
+
+static PyObject *CPyDict_KeysView(PyObject *dict) {
+    if (PyDict_CheckExact(dict)){
+        return _CPyDictView_New(dict, &PyDictKeys_Type);
+    }
+    return PyObject_CallMethod(dict, "keys", NULL);
+}
+
+static PyObject *CPyDict_ValuesView(PyObject *dict) {
+    if (PyDict_CheckExact(dict)){
+        return _CPyDictView_New(dict, &PyDictValues_Type);
+    }
+    return PyObject_CallMethod(dict, "values", NULL);
+}
+
+static PyObject *CPyDict_ItemsView(PyObject *dict) {
+    if (PyDict_CheckExact(dict)){
+        return _CPyDictView_New(dict, &PyDictItems_Type);
+    }
+    return PyObject_CallMethod(dict, "items", NULL);
+}
+
+static PyObject *CPyDict_Keys(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        return PyDict_Keys(dict);
+    }
+    // Inline generic fallback logic to also return a list.
+    PyObject *list = PyList_New(0);
+    PyObject *view = PyObject_CallMethod(dict, "keys", NULL);
+    if (view == NULL) {
+        return NULL;
+    }
+    PyObject *res = _PyList_Extend((PyListObject *)list, view);
+    Py_DECREF(view);
+    if (res == NULL) {
+        return NULL;
+    }
+    Py_DECREF(res);
+    return list;
+}
+
+static PyObject *CPyDict_Values(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        return PyDict_Values(dict);
+    }
+    // Inline generic fallback logic to also return a list.
+    PyObject *list = PyList_New(0);
+    PyObject *view = PyObject_CallMethod(dict, "values", NULL);
+    if (view == NULL) {
+        return NULL;
+    }
+    PyObject *res = _PyList_Extend((PyListObject *)list, view);
+    Py_DECREF(view);
+    if (res == NULL) {
+        return NULL;
+    }
+    Py_DECREF(res);
+    return list;
+}
+
+static PyObject *CPyDict_Items(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        return PyDict_Items(dict);
+    }
+    // Inline generic fallback logic to also return a list.
+    PyObject *list = PyList_New(0);
+    PyObject *view = PyObject_CallMethod(dict, "items", NULL);
+    if (view == NULL) {
+        return NULL;
+    }
+    PyObject *res = _PyList_Extend((PyListObject *)list, view);
+    Py_DECREF(view);
+    if (res == NULL) {
+        return NULL;
+    }
+    Py_DECREF(res);
+    return list;
+}
+
+static PyObject *CPyDict_GetKeysIter(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        // Return dict itself to indicate we can use fast path instead.
+        Py_INCREF(dict);
+        return dict;
+    }
+    return PyObject_GetIter(dict);
+}
+
+static PyObject *CPyDict_GetItemsIter(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        // Return dict itself to indicate we can use fast path instead.
+        Py_INCREF(dict);
+        return dict;
+    }
+    PyObject *view = PyObject_CallMethod(dict, "items", NULL);
+    if (view == NULL) {
+        return NULL;
+    }
+    PyObject *iter = PyObject_GetIter(view);
+    Py_DECREF(view);
+    return iter;
+}
+
+static PyObject *CPyDict_GetValuesIter(PyObject *dict) {
+    if (PyDict_CheckExact(dict)) {
+        // Return dict itself to indicate we can use fast path instead.
+        Py_INCREF(dict);
+        return dict;
+    }
+    PyObject *view = PyObject_CallMethod(dict, "values", NULL);
+    if (view == NULL) {
+        return NULL;
+    }
+    PyObject *iter = PyObject_GetIter(view);
+    Py_DECREF(view);
+    return iter;
+}
+
+// Our return tuple wrapper for dictionary iteration helper.
+#ifndef MYPYC_DECLARED_tuple_T3CIO
+#define MYPYC_DECLARED_tuple_T3CIO
+typedef struct tuple_T3CIO {
+    char f0;  // Should continue?
+    CPyTagged f1;  // Last dict offset
+    PyObject *f2;  // Next dictionary key or value
+} tuple_T3CIO;
+static tuple_T3CIO tuple_undefined_T3CIO = { 2, CPY_INT_TAG, NULL };
+#endif
+
+// Same as above but for both key and value.
+#ifndef MYPYC_DECLARED_tuple_T4CIOO
+#define MYPYC_DECLARED_tuple_T4CIOO
+typedef struct tuple_T4CIOO {
+    char f0;  // Should continue?
+    CPyTagged f1;  // Last dict offset
+    PyObject *f2;  // Next dictionary key
+    PyObject *f3;  // Next dictionary value
+} tuple_T4CIOO;
+static tuple_T4CIOO tuple_undefined_T4CIOO = { 2, CPY_INT_TAG, NULL, NULL };
+#endif
+
+static void _CPyDict_FromNext(tuple_T3CIO *ret, PyObject *dict_iter) {
+    // Get next item from iterator and set "should continue" flag.
+    ret->f2 = PyIter_Next(dict_iter);
+    if (ret->f2 == NULL) {
+        ret->f0 = 0;
+        Py_INCREF(Py_None);
+        ret->f2 = Py_None;
+    } else {
+        ret->f0 = 1;
+    }
+}
+
+// Helpers for fast dictionary iteration, return a single tuple
+// instead of writing to multiple registers, for exact dicts use
+// the fast path, and fall back to generic iterator logic for subclasses.
+static tuple_T3CIO CPyDict_NextKey(PyObject *dict_or_iter, CPyTagged offset) {
+    tuple_T3CIO ret;
+    Py_ssize_t py_offset = CPyTagged_AsSsize_t(offset);
+    PyObject *dummy;
+
+    if (PyDict_CheckExact(dict_or_iter)) {
+        ret.f0 = PyDict_Next(dict_or_iter, &py_offset, &ret.f2, &dummy);
+        if (ret.f0) {
+            ret.f1 = CPyTagged_FromSsize_t(py_offset);
+        } else {
+            // Set key to None, so mypyc can manage refcounts.
+            ret.f1 = 0;
+            ret.f2 = Py_None;
+        }
+        // PyDict_Next() returns borrowed references.
+        Py_INCREF(ret.f2);
+    } else {
+        // offset is dummy in this case, just use the old value.
+        ret.f1 = offset;
+        _CPyDict_FromNext(&ret, dict_or_iter);
+    }
+    return ret;
+}
+
+static tuple_T3CIO CPyDict_NextValue(PyObject *dict_or_iter, CPyTagged offset) {
+    tuple_T3CIO ret;
+    Py_ssize_t py_offset = CPyTagged_AsSsize_t(offset);
+    PyObject *dummy;
+
+    if (PyDict_CheckExact(dict_or_iter)) {
+        ret.f0 = PyDict_Next(dict_or_iter, &py_offset, &dummy, &ret.f2);
+        if (ret.f0) {
+            ret.f1 = CPyTagged_FromSsize_t(py_offset);
+        } else {
+            // Set value to None, so mypyc can manage refcounts.
+            ret.f1 = 0;
+            ret.f2 = Py_None;
+        }
+        // PyDict_Next() returns borrowed references.
+        Py_INCREF(ret.f2);
+    } else {
+        // offset is dummy in this case, just use the old value.
+        ret.f1 = offset;
+        _CPyDict_FromNext(&ret, dict_or_iter);
+    }
+    return ret;
+}
+
+static tuple_T4CIOO CPyDict_NextItem(PyObject *dict_or_iter, CPyTagged offset) {
+    tuple_T4CIOO ret;
+    Py_ssize_t py_offset = CPyTagged_AsSsize_t(offset);
+
+    if (PyDict_CheckExact(dict_or_iter)) {
+        ret.f0 = PyDict_Next(dict_or_iter, &py_offset, &ret.f2, &ret.f3);
+        if (ret.f0) {
+            ret.f1 = CPyTagged_FromSsize_t(py_offset);
+        } else {
+            // Set key and value to None, so mypyc can manage refcounts.
+            ret.f1 = 0;
+            ret.f2 = Py_None;
+            ret.f3 = Py_None;
+        }
+    } else {
+        ret.f1 = offset;
+        PyObject *item = PyIter_Next(dict_or_iter);
+        if (item == NULL || !PyTuple_Check(item) || PyTuple_GET_SIZE(item) != 2) {
+            if (item != NULL) {
+                PyErr_SetString(PyExc_TypeError, "a tuple of length 2 expected");
+            }
+            ret.f0 = 0;
+            ret.f2 = Py_None;
+            ret.f3 = Py_None;
+        } else {
+            ret.f0 = 1;
+            ret.f2 = PyTuple_GET_ITEM(item, 0);
+            ret.f3 = PyTuple_GET_ITEM(item, 1);
+            Py_DECREF(item);
+        }
+    }
+    // PyDict_Next() returns borrowed references.
+    Py_INCREF(ret.f2);
+    Py_INCREF(ret.f3);
+    return ret;
+}
+
+// Check that dictionary didn't change size during iteration.
+static inline char CPyDict_CheckSize(PyObject *dict, CPyTagged size) {
+    if (!PyDict_CheckExact(dict)) {
+        // Dict subclasses will be checked by Python runtime.
+        return 1;
+    }
+    Py_ssize_t py_size = CPyTagged_AsSsize_t(size);
+    Py_ssize_t dict_size = PyDict_Size(dict);
+    if (py_size != dict_size) {
+        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
+        return 0;
+    }
+    return 1;
+}
+
 void CPy_Init(void);
 
 
@@ -1302,7 +1671,7 @@ static int CPy_YieldFromErrorHandle(PyObject *iter, PyObject **outp)
     } else {
         _m = _PyObject_GetAttrId(iter, &PyId_throw);
         if (_m) {
-            CPy_GetExcInfo(&type, &value, &traceback);
+            _CPy_GetExcInfo(&type, &value, &traceback);
             res = PyObject_CallFunctionObjArgs(_m, type, value, traceback, NULL);
             Py_DECREF(type);
             Py_DECREF(value);
